#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils, views
from w3bsite.classes import defaults as _defaults_

# the authentication requests.
class Requests(_defaults_.Defaults):
	def __init__(self, 
		# passed Website.x objects.
		defaults=None,
	):
		# defaults.
		_defaults_.Defaults.__init__(self)
		self.assign(defaults.dict())

		# urlpatterns.
		send_code = self.SendCode(defaults=defaults)
		self.urls = views.build_urls([
			self.SignUp(defaults=defaults, send_code=send_code),
			self.SignIn(defaults=defaults),
			send_code,
			self.ResetPassword(defaults=defaults),
			self.Activate(defaults=defaults),
		])

	# sign up.
	class SignUp(views.Request):
		def __init__(self, defaults=None, send_code=None):
			views.Request.__init__(self, "requests/authentication/", "signup")
			self.assign(defaults.dict())
			self.send_code = send_code
		def view(self, request):

			# permission denied.
			#return self.permission_denied()

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# check signup rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="signup", limit=3, reset_minutes=3600*24, increment=False)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"username",
				"email",
				"password",
				"verify_password",])
			if not response.success: return self.response(response)
			optional_parameters, _ = self.parameters.get(request, {
				"name":None,
			})

			# make request.
			response = self.users.create(
				email=parameters["email"],
				username=parameters["username"],
				password=parameters["password"],
				verify_password=parameters["verify_password"],
				name=optional_parameters["name"],)
			if not response.success: return self.response(response)

			# increment.
			_response_ = self.rate_limit.increment(ip=utils.get_client_ip(request), mode="signup")
			if not _response_.success: return _response_


			# send activation code.
			_response_ = self.send_code.send_code(username=parameters["username"], mode="activation", request=request)
			if not _response_.success: return self.response(_response_)

			# return create response.
			if response.success:
				del response["user"]
			return self.response(response)

	# sign in.
	class SignIn(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "signin")
			self.assign(defaults.dict())
		def view(self, request):

			# permission denied.
			#return self.permission_denied()

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# check signup rate limit.
			#response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="signin", limit=5, reset_minutes=10, increment=True)
			#if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"username",
				"password",])
			if not response.success: return self.response(response)
			optional_parameters, _ = self.parameters.get(request, {
				"code":None,
			})

			# html
			html = ""
			if self._2fa:

				# set mode.
				code = Integer(0).generate(length=6)
				title = "Sign In - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/authentication.html"
				ip = utils.__get_client_ip__(request)
				html = str(File(path, load=True).data).format(
					# domain info.
					domain=self.https_domain,
					# code.
					code=code,
					# request info.
					ip=ip,
					timestamp=Date().seconds_timestamp,
					# user info.
					email=parameters.username,
					# colors.
					white=self.template_data["COLORS"]["white"],
					grey=self.template_data["COLORS"]["grey"],
					blue=self.template_data["COLORS"]["blue"],
					purple=self.template_data["COLORS"]["purple"],
					red=self.template_data["COLORS"]["red"],
					pink=self.template_data["COLORS"]["pink"],
					orange=self.template_data["COLORS"]["orange"],
					green=self.template_data["COLORS"]["green"],
					darkest=self.template_data["COLORS"]["darkest"],
					darker=self.template_data["COLORS"]["darker"],
					dark=self.template_data["COLORS"]["dark"],
					background=self.template_data["COLORS"]["background"],
					widgets=self.template_data["COLORS"]["widgets"],
					widgets_reversed=self.template_data["COLORS"]["widgets_reversed"],
					text=self.template_data["COLORS"]["text"],
					text_reversed=self.template_data["COLORS"]["text_reversed"],
					)

			# make request.
			return self.response(self.users.authenticate(
				username=parameters["username"],
				password=parameters["password"],
				_2fa_code=optional_parameters["code"],
				_2fa=self._2fa,
				html=html,
				request=request,))

	# send reset password email.
	class SendCode(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "send_code", )
			self.assign(defaults.dict())
		def view(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# check 2fa rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="2fa", limit=15, reset_minutes=3600, increment=True)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"username",
				"mode",])
			if not response.success: return self.response(response)
			if parameters["username"] in ["undefined", "null", "None", "none", ""]:
				return self.error("Refresh the web page.")

			# make request.
			return self.response(self.send_code(
				username=parameters["username"],
				mode=parameters['mode'],
				request=request,))

			#
		def send_code(self, username=None, mode=None, request=None):

			# check library.
			#if not Files.exists("templates/mail"): os.mkdir("templates/mail")
			#response = utils.__check_package_files__([
			#	["templates/mail/reset_password.html","w3bsite","/classes/requests/mail/reset_password.html"],
			#	["templates/mail/authentication.html","w3bsite","/classes/requests/mail/authentication.html"],
			#	["templates/mail/activation.html","w3bsite","/classes/requests/mail/activation.html"],
			#])
			#if not response.success: return response

			# set mode.
			code = Integer(0).generate(length=6)
			if mode == "reset_password":
				title = "Reset Password - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/reset_password.html"
			elif mode == "authentication":
				title = "Sign In - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/authentication.html"
			elif mode == "activation":
				title = "Account Activation - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/activation.html"
			else:
				return Response.error("Selected an invalid mode.")

			# parse html.
			ip = utils.__get_client_ip__(request)
			html = str(File(path, load=True).data).format(
				# domain info.
				domain=self.https_domain,
				# code.
				code=code,
				# request info.
				ip=ip,
				timestamp=Date().seconds_timestamp,
				# user info.
				email=username,
				# colors.
				white=self.template_data["COLORS"]["white"],
				grey=self.template_data["COLORS"]["grey"],
				blue=self.template_data["COLORS"]["blue"],
				purple=self.template_data["COLORS"]["purple"],
				red=self.template_data["COLORS"]["red"],
				pink=self.template_data["COLORS"]["pink"],
				orange=self.template_data["COLORS"]["orange"],
				green=self.template_data["COLORS"]["green"],
				darkest=self.template_data["COLORS"]["darkest"],
				darker=self.template_data["COLORS"]["darker"],
				dark=self.template_data["COLORS"]["dark"],
				background=self.template_data["COLORS"]["background"],
				widgets=self.template_data["COLORS"]["widgets"],
				widgets_reversed=self.template_data["COLORS"]["widgets_reversed"],
				text=self.template_data["COLORS"]["text"],
				text_reversed=self.template_data["COLORS"]["text_reversed"],
				)

			# make request.
			return self.users.send_code(
				username=username,
				mode=mode,
				ip=ip,
				code=code,
				html=html,
				title=title,)

	# reset password.
	class ResetPassword(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "reset_password",)
			self.assign(defaults.dict())
		def view(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"code",
				"email",
				"password",
				"verify_password",])
			if not response.success: return self.response(response)
			if parameters["email"] in ["undefined", "null", "None", "none", ""]:
				return self.error("Refresh the web page.")

			# make request.
			response = self.users.verify_code(
				email=parameters["email"],
				code=parameters["code"],
				mode="reset_password",)
			if not response.success: return self.response(response)

			# make request.
			response = self.users.update(
				email=parameters["email"], # for id.
				password=parameters["password"],
				verify_password=parameters["verify_password"],)
			if response.success: 
				response["message"] = f"Succesfully resetted the password of user [{user.email}]."
				_response_ = self.users.save_password(email=user.email, username=user.username, password=parameters["password"])
				if not _response_.success: return self.response(_response_)
			return self.response(response)

	# activate account.
	class Activate(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "activate")
			self.assign(defaults.dict())
		def view(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"code",
				"email",])
			if not response.success: return self.response(response)
			if parameters["email"] in ["undefined", "null", "None", "none", ""]:
				return self.error("Refresh the web page.")

			# make request.
			response = self.users.verify_code(
				email=parameters["email"],
				code=parameters["code"],
				mode="activation",)
			if not response.success: return self.response(response)

			# make request.
			response = self.users.update(
				email=parameters["email"],
				email_verified=True,)
			if response.success: response["message"] = f"Succesfully activated the account user [{user.email}]."
			return self.response(response)

	# is authenticated.
	class Authenticated(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "activate")
			self.assign(defaults.dict())
		def view(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# handler.
			return self.response("Successfully checked if the user is authenticated.", {
				"authenticated":request.user.username != None and request.user.is_authenticated == True,
				"username":request.user.username,
				"email":request.user.email,
			})


# user wide cache to share variables accross tty sessions since default env variables is instable.
import syst3m
from dev0s import *
if not Files.exists(f"{Defaults.vars.home}/.w3bsite/"): 
	Files.create(path=f"{Defaults.vars.home}/.w3bsite/", directory=True)
cache = syst3m.cache.Cache(
	path=f"{Defaults.vars.home}/.w3bsite/.cache/")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *

# the synchronize users thread (synchs once per hour).
class Synchronize(threading.Thread):
	def __init__(self, 
		# pass the Website.users object.
		users=None,
		# show the daemon logs.
		silent=False,
		# the synchronize interval in minutes (default: synchronizes once per 60 min).
		synchronize_interval=60,
		# the sleep interval in minutes (default: checks synchronize interval once per 15 min).
		sleep_interval=15,
	):
		threading.Thread.__init__(self)
		self.users = users
		self.silent = silent
		self.synchronize_interval = synchronize_interval
		self.sleep_interval = sleep_interval
		self.monthly_requests, self.monthly_costs = self.get_costs(self.synchronize_interval, len(self.users.iterate(filter="users")))
		monthly_requests, costs = self.get_costs(self.synchronize_interval, 10000)
		print(f"Monthly daemon requests: {int(self.monthly_requests)}")
		print(f"Monthly daemon costs: € {self.monthly_costs:.2f}")
		print(f"Monthly daemon costs per 10.000 users: € {costs:.2f}")
	def run(self):
		last = None
		while True:
			date, new = Date(), False
			if last == None: new = True
			else:
				increased = date.increase(last, minutes=self.synchronize_interval)
				new = date.compare(increased, date.timestamp) in ["present", "past"]
			if new:
				last = date.timestamp
				response = self.users.synchronize()
				Response.log(response, log_level=-1, save_errors=True)
			time.sleep(self.sleep_interval*60)
	def get_costs(self, synchronize_interval=None, total_users=None):
		monthly_requests = ( total_users * (( 60 * 24 ) / synchronize_interval )) * 30.5
		monthly_costs = monthly_requests * (0.06/100000)
		return monthly_requests, monthly_costs


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
import requests, sys, os, ast, json, glob, platform, time
from django.shortcuts import render
from django.http import HttpResponse, HttpResponseNotAllowed, JsonResponse

# inc imports.
import syst3m, w3bsite
from dev0s import *

# source.
SOURCE_NAME = ALIAS = "api.vandenberghinc.com"
SOURCE_PATH = Defaults.source_path(__file__, back=3)
OS = Defaults.operating_system(supported=["linux", "macos"])
#Defaults.alias(alias=ALIAS, executable=SOURCE_PATH, sudo=True)

# production settings.
# do not deploy to heroku with production disabled.
PRODUCTION = Environment.get_boolean("PRODUCTION", default=True)
MAINTENANCE = Environment.get_boolean("MAINTENANCE", default=True)

# website.
website = w3bsite.Website(
	serialized=f"__defaults__/env/website",
	django=True,
	maintenance=True,
	root=SOURCE_PATH, 
	production=PRODUCTION,)

# synchronize users.
if Environment.get_boolean("MIGRATIONS") == False:
	response = website.users.synchronize()
	if not response.success: raise ValueError(response['error'])

# colors.
COLORS = {
	"white":"#FAFAFA",
	"light_white":"#E9F0FD",
	"grey":"#E5E5E5",
	"light_grey":"#D6D6D6",
	"dark_grey":"#424242",
	"blue":"#5A8FE6",
	"purple":"#323B83",#"#B32FCA",
	#"purple":"#9B00AA",
	"red":"#FD304E",
	"pink":"#F62B7D",
	"orange":"#FF8800",
	"green":"#006633",
	"darkest":"#1F2227",
	"darker": "#20242A",
	"dark": "#262B30",
	# background color.
	"topbar":"#FAFAFA",#1F2227", #"#FAFAFA",
	"background":"#323B83",#"#E7E9EF", #"#FAFAFA",
	"topbar_darkmode":"#1F2227",#1F2227", #"#FAFAFA",
	"background_darkmode":"#1F2227",#"#E7E9EF", #"#FAFAFA",
	# elements.
	"widgets":"#FAFAFA",
	"widgets_reversed":"#323B83",#"#1F2227",
	"widgets_darkmode":"#20242A",
	"widgets_reversed_darkmode":"#323B83",#"#1F2227",
	# text.
	"text":"#1F2227",
	"text_reversed":"#FAFAFA",
	"text_darkmode":"#FAFAFA",
	"text_reversed_darkmode":"#FAFAFA",
	# input & textareas.
	"input_txt":"#6C6B6D",
	"input_txt_reversed":"#FAFAFA",
	"input_bg":"#E9F0FD", #"#FAFAFA", 
	"input_bg_reversed":"#323B83",
	
	# buttons.
	"button_txt":"#FAFAFA",
	"button_txt_reversed":"#1F2227",
	"button_bg":"#323B83",
	"button_bg_reversed":"#FAFAFA",
	# custom colors.
	# ...
}

# template data.
website.template_data["COLORS"] = COLORS

# database.
DATABASE = website.database
if not Files.exists(DATABASE): os.system(f"sudo mkdir {DATABASE} && sudo chown {Defaults.vars.user}:{Defaults.vars.group} {DATABASE}")
if not Files.exists(f"{DATABASE}/data/"): os.mkdir(f"{DATABASE}/data/") # for database.
if not Files.exists(f"{DATABASE}/packages/"): os.mkdir(f"{DATABASE}/packages/")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils

# the deployment object class.
class Deployment(Object):
	# does not use Defaults so also accessable without Website init.
	# deploy the website on a local ubuntu machine.
	def __init__(self,
		# the root path.
		root=None,
		# the library path.
		library=None,
		# the website name.
		name=None,
		# the domain.
		domain=None,
		# the database path.
		database=None,
		# the remote.
		remote=None,
		# the vps ip (if remote is vps else leave default).
		vps_ip=None,
		vps_username=None,
		# the organization's email.
		email=None,
		country_code="NL",
		province="Amsterdam",
		city="Amsterdam",
		organization=None,
		organization_unit="IT",
		# the organization info.
		# objects.
		namecheap=None,
	):	

		# defaults.
		Object.__init__(self, traceback="w3bsite.Website.deployment",)

		# attributes.
		self.vps_ip = vps_ip
		self.vps_username = vps_username
		self.root = root
		self.library = library
		self.domain = domain
		self.database = database
		self.remote = remote
		self.name = name

		self.country_code = country_code
		self.province = province
		self.city = city
		self.organization = organization
		self.organization_unit = organization_unit
		self.email = email

		# vars.
		self.live = gfp.clean(self.library, remove_double_slash=True, remove_last_slash=True) in gfp.clean(self.root, remove_double_slash=True, remove_last_slash=True)

		# objects.
		self.namecheap = namecheap

		#
	# live functions.
	def start(self):

		# execute.
		command = ""
		for i in ["gunicorn.socket", "gunicorn", "nginx"]:
			command += f"sudo systemctl start {i} &&"
		command = command[:-3]
		response = Code.execute(command)
		if not response.success: return response
		output = response.output
		if output in ["", "\n"]:
			return Response.success(f"Successfully started {self.name}")
		else:
			return Response.error(f"Failed to started {self.name};\n{output}")

		#
	def stop(self):

		# execute.
		command = ""
		for i in ["gunicorn.socket", "gunicorn", "nginx"]:
			command += f"sudo systemctl stop {i} &&"
		command = command[:-3]
		response = Code.execute(command)
		if not response.success: return response
		output = response.output
		if output in ["", "\n"]:
			return Response.success(f"Successfully stopped {self.name}")
		else:
			return Response.error(f"Failed to stopped {self.name};\n{output}")

		#
	def restart(self):

		# execute.
		command = ""
		for i in ["gunicorn.socket", "gunicorn", "nginx"]:
			command += f"sudo systemctl restart {i} &&"
		command = command[:-3]
		response = Code.execute(command)
		if not response.success: return response
		output = response.output
		if output in ["", "\n"]:
			return Response.success(f"Successfully restarted {self.name}")
		else:
			return Response.error(f"Failed to restarted {self.name};\n{output}")

		#
	def status(self):

		# execute.
		response = Code.execute("sudo systemctl status gunicorn > /tmp/status && cat /tmp/status && rm -fr /tmp/status")
		if not response.success: return response
		status = response.output
		return Response.success(f"Successfully parsed the status of {self.name}.", {
			"status":status,
		})

		#
	def reset_logs(self):

		# execute.
		Files.save(f"{self.database}/logs/logs", "")
		Files.save(f"{self.database}/logs/errors", "")
		Files.save(f"{self.database}/logs/nginx", "")
		Files.save(f"{self.database}/logs/nginx.debug", "")
		return Response.success("Successfully resetted the logs.")

		#
	def tail(self, nginx=False, debug=False):
		
		# execute.
		if nginx:
			if debug:
				data = Files.load(f"/{self.database}/logs/nginx.debug")
			else:
				data = Files.load(f"/{self.database}/logs/nginx")
		else:
			if debug:
				data = Files.load(f"{self.database}/logs/logs")
			else:
				data = Files.load(f"{self.database}/logs/errors")
		
		# handler.
		return Response.success(f"Successfully tailed the {self.name} logs.", {
			"logs":data,
		})

		#

	# configure also for remove:vps & remote:local.
	def configure(self, reinstall=False, log_level=0, loader=None):
		
		# check arguments.
		if self.remote in ["vps"]:
			response = Response.parameters.check(
				traceback=self.__traceback__(function="configure"),
				parameters={
					"vps_ip":self.vps_ip,
					"vps_username":self.vps_username,
				})
			if not response.success: return response
			username = self.vps_username
		else:
			username = Defaults.vars.user
		
		# configure before loader.
		if self.live:
			if reinstall:
				os.system(f"rm -fr {self.database}/tls/dhparam.pem")
			if not Files.exists(f"{self.database}/tls/dhparam.pem"):
				if loader != None: loader.hold()
				tmp = "/tmp/dhparam.pem"
				os.system(f"sudo openssl dhparam -out {tmp} 4096 && sudo chown {Defaults.vars.user}:{Defaults.vars.group} {tmp} && mv {tmp} {self.database}/tls/dhparam.pem && sudo chown {Defaults.vars.user}:{Defaults.vars.group} {self.database}/tls/dhparam.pem")
				if loader != None: loader.release()

		# loader.
		if log_level >= 0: loader = Console.Loader(f"Configuring deployment of website {self.domain} ...")

		# check remote.
		if self.remote in ["heroku"]:
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"You can not execute function <Website.deployment.configure> with remote [{self.remote}].")

		# os.
		if OS not in ["macos", "linux"]: 
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Unsupported operating system [{OS}].")

		# requirements.
		if not Files.exists(f"{self.root}/requirements"): os.mkdir(f"{self.root}/requirements")
		if not Files.exists(f"{self.root}/requirements/requirements.pip"): 
			Files.save(f"{self.root}/requirements/requirements.pip", "wheel\nuwsgi\ngunicorn\nwhitenoise\ndjango\npsycopg2-binary\nsyst3m\nw3bsite")
		if not Files.exists(f"{self.root}/requirements/installer"): 
			os.system(f"cp {SOURCE_PATH}/example/requirements/installer {self.root}/requirements/installer && chmod +x {self.root}/requirements/installer")

		# favicon.
		if self.live:
			if not Files.exists(f"{self.root}/__defaults__/static/favicon.ico"):
				response = Code.execute(f"curl -s https://raw.githubusercontent.com/vandenberghinc/public-storage/master/w3bsite/favicon.ico -o {self.root}/__defaults__/static/favicon.ico")
				if not response.success: return response
				output = response.output

		# tls.
		if self.live:
			if not Files.exists(f"{self.database}/tls/server.key") and not Files.exists(f"{self.database}/tls/server.crt"):
				response = self.generate_tls(log_level=log_level)
				if not response.success: 
					if log_level >= 0: loader.stop(success=False)
					return response

		# database.
		if self.live:
			if not Files.exists(self.database): 
				os.system(f"sudo mkdir -p {self.database} && sudo chown {Defaults.vars.user}:{Defaults.vars.group} {self.database} && sudo chmod 770 {self.database}")
			if not Files.exists(f"{self.database}/logs"): os.mkdir(f"{self.database}/logs")

		# deployment.
		if not Files.exists(f"{self.root}/deployment"): os.mkdir(f"{self.root}/deployment")
		clean_root = gfp.clean(self.library, remove_last_slash=True, remove_double_slash=True) # <== note the library change instead of root.
		replacements = {
			"***ROOT***":clean_root, 
			"***USER***":username, 
			"***DOMAIN***":self.domain,
			"***DATABASE***":self.database,
			"***USER***":Defaults.vars.user,
			"***WEBSITE_BASE***":gfp.base(SOURCE_PATH),
		}
		for path in Files.Directory(path=f"{SOURCE_PATH}/classes/deployment/lib/").paths():
			name = FilePath(path).name()
			try:
				data = Files.load(path)
			except FileNotFoundError:
				data = ""
			new_data = str(data)
			for key,value in replacements.items():
				new_data = new_data.replace(key, value)
			Files.save(f"{self.root}/deployment/{name}", new_data)
		
		# success.
		if log_level >= 0: loader.stop()
		return Response.success(f"Successfully configured the deployment of domain {self.domain}.")

		#
	# deploy is for remote:local only. 
	def deploy(self, code_update=False, reinstall=False, log_level=0):
			
		# loader.
		loader = None
		if log_level >= 0: loader = Console.Loader(f"Deploying domain {self.domain} ...")

		# check namecheap domain.
		if not (self.remote in ["vps"] and self.live):
			response = self.namecheap.check_domain(self.namecheap.post_domain)
			if response.error != None: 
				if log_level >= 0: loader.stop(success=False)
				Response.log(response=response, log_level=log_level)
				return response
			elif not response["exists"]:
				if log_level >= 0: loader.stop(success=False)
				return Response.error(f"Specified domain [{self.namecheap.post_domain}] is not owned by namecheap user [{self.namecheap.username}].", log_level=log_level)

		# check remote.
		if self.remote in ["vps"] and not self.live:
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"You can not execute function <Website.deployment.deploy> with remote [{self.remote}].")

		# os.
		if OS not in ["linux"]: 
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Unsupported operating system [{OS}].")
		
		# configure.
		response = self.configure(reinstall=reinstall, log_level=log_level, loader=loader)
		if not response.success: 
			if log_level >= 0: loader.stop(quiet=True)
			return response

		# check tls domain.
		if not Files.exists(f"{self.database}/tls/.domain"): Files.save(f"{self.database}/tls/.domain", self.domain)
		tls_domain = Files.load(f"{self.database}/tls/.domain").replace('\n',"")
		if tls_domain != self.domain:
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"TLS Certificate mis match. Installed tls certificate [{self.database}/tls] is linked to domain {tls_domain}, not specified domain {self.domain}.", log_level=0)
		
		# checks.
		if not Files.exists(f"{self.database}/tls/server.key") or not Files.exists(f"{self.database}/tls/server.crt"):
			if log_level >= 0: loader.stop(success=False)
			return Response.error("No tls certificate exists.\nExecute the following command to generate a tls certificate:\n$ ./website.py --generate-tls", log_level=log_level)
		#if not Files.exists(f"{self.database}/tls/signed.server.crt"):
		#	if log_level >= 0: loader.stop(success=False)
		#	return Response.error("No activated tls certificate exists. \nExecute the following command to activate the generated tls certificate:\n$ ./website.py --activate-tls", log_level=log_level)
		#if not Files.exists(f"{self.database}/tls/server.ca-bundle"):
		#	if log_level >= 0: loader.stop(success=False)
		#	return Response.error("No bundled tls certificate exists. \nDownload the signed certificate send to your email, extr the zip to a directory and execute \n$ ./website.py --bundle-tls /path/to/extracted/directory/", log_level=log_level)
		
		# arguments.
		arguments = ""
		if code_update: arguments += " --code-update"
		if reinstall: arguments += " --reinstall"
		
		# execute & handle.
		Files.chmod(f"{self.root}/deployment/installer", "+x")
		response = Code.execute(f"bash {self.root}/deployment/installer{arguments}")
		if not response.success:
			#if "Error:" in output or ("nginx: the configuration file /etc/nginx/nginx.conf syntax is ok" not in output and "nginx: configuration file /etc/nginx/nginx.conf test is successful" not in output): #"Successfully deployed domain " not in output
			print(response.output)
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Failed to deploy website {self.domain}.", log_level=log_level)
		else:
			if log_level >= 0: loader.stop()
			return Response.success(f"Successfully deployed domain https://{self.domain}.", log_level=log_level)

		#
	def generate_tls(self, log_level=0):
		# https://devcenter.heroku.com/articles/acquiring-an-ssl-certificate

		# check base.
		base = f"{self.root}/__defaults__/"
		if not Files.exists(base): os.mkdir(base)
		base = f"{self.database}/tls"
		if not Files.exists(base): os.mkdir(base)

		# check duplicate.
		if Files.exists(f"{self.database}/tls/server.key") or Files.exists(f"{self.database}/tls/server.crt"):
			return Response.error("The tls certificate already exists.", log_level=log_level)

		# generate.
		if log_level >= 0: loader = Console.Loader("Generating a tls certificate ...")
		response = Code.execute(f"""
			cd {base}

			# Generate a passphrase
			openssl rand -base64 48 > passphrase.txt

			# Generate a Private Key
			openssl genrsa -aes128 -passout file:passphrase.txt -out server.key 4096 # 2048

			# Generate a CSR (Certificate Signing Request)
			openssl req -new -passin file:passphrase.txt -key server.key -out server.csr \
			    -subj "/C={self.country_code}/ST={self.province}/L={self.city}/O={self.organization}/OU={self.organization_unit}/CN={self.domain}/emailAddress={self.email}"

			# Remove Passphrase from Key
			cp server.key server.pass.key
			openssl rsa -in server.pass.key -passin file:passphrase.txt -out server.key

			# Do not self sign.
			# Generating a Self-Signed Certificate for 100 years
			openssl x509 -req -sha256 -days 36500 -in server.csr -signkey server.key -out server.crt

		""")
		if not response.success: return response
		output = response.output

		# handler.
		if not Files.exists(f"{self.database}/tls/server.key") or not Files.exists(f"{self.database}/tls/server.crt"):
			if log_level >= 0: loader.stop(success=False)
			os.system(f"rm -fr {base}")
			return Response.error(f"Failed to generate a tls certificate.", log_level=log_level)
		else:
			if log_level >= 0: loader.stop()
			Files.save(f"{self.database}/tls/.domain", self.domain)
			return Response.success(f"Successfully generated a tls certificate.", log_level=log_level)

		#
	def activate_tls(self, log_level=0):

		# check existsance.
		if not Files.exists(f"{self.database}/tls/server.key") or not Files.exists(f"{self.database}/tls/server.crt"):
			return Response.error("No generated tls certificate exists.", log_level=log_level)

		# check duplicate.
		if Files.exists(f"{self.database}/tls/signed.server.key"):
			return Response.error("A signed tls certificate already exists.", log_level=log_level)

		response = self.namecheap.get_tls()
		if response.error != None: 
			Response.log(response=response, log_level=log_level)
			return response
		tls, certificates = response["tls"], response["certificates"]
		
		# check no certificates.
		if certificates == 0:

			# create tls.
			loader = Console.Loader("Purchasing a namecheap tls certificate ...")
			response = self.namecheap.create_tls(
				# the expiration years.
				years=2,
				# the tls type.
				type="PositiveSSL",)
			Response.log(response=response, log_level=log_level)
			if response.error != None: 
				loader.stop(success=False)
				return response	
			loader.stop()
			
			# get tls again.
			response = self.namecheap.get_tls()
			if response.error != None: 
				loader.stop(success=False)
				Response.log(response=response, log_level=log_level)
				return response
			loader.stop()
			tls, certificates = response["tls"], response["certificates"]

		# check root domain tls.
		tls_activated = False
		for certificate_id, info in tls.items():
			if info["host_name"] == self.domain:
				tls_activated = True
				break
		if not tls_activated:

			# get new purchase certficiate.
			id = None
			for certificate_id, info in tls.items():
				if info["status"] == "newpurchase":
					id = certificate_id
					break

			# create certficiate on no new purchase found.
			if id == None:
				loader = Console.Loader("Purchasing a namecheap tls certificate ...")
				response = self.namecheap.create_tls(
					# the expiration years.
					years=2,
					# the tls type.
					type="PositiveSSL",)
				Response.log(response=response, log_level=log_level)
				if response.error != None: 
					loader.stop(success=False)
					return response	
				loader.stop()
				id = response["certificate_id"]

			# activate tls for root domain.
			loader = Console.Loader("Activating tls certificate ...")
			response = self.namecheap.activate_tls(
				# the certificate's id.
				certificate_id=id,)
			loader.stop(success=response["success"])
			Response.log(response=response, log_level=log_level)
			if response.error != None: return response
			
		# handlers.
		return Response.success(f"Successfully activated the tls certificate of domain [{self.domain}]. Wait for the CA to send you a .zip file with your signed certificate. Extract the zip & bundle the certificate with: $ ./website.py --bundle-tls /path/to/extracted/directory/certificate/", log_level=log_level)

		#
	def bundle_tls(self, directory, log_level=0):
		
		# check dir.
		if not Files.exists(directory):
			return Response.error(f"Specified directory [{directory}] does not exist.", log_level=log_level)
		if ".zip" in directory:
			return Response.error(f"Specified directory [{directory}] is zip format, extract the zip first.", log_level=log_level)
		if not os.path.isdir(directory):
			return Response.error(f"Specified directory [{directory}] is not a directory.", log_level=log_level)
		
		# move x.crt to server.crt
		if not Files.exists(f"{directory}/server.crt") and not Files.exists(f'{directory}/{self.domain.replace(".","_")}.crt'):
			return Response.success(f'You must rename the [{directory}/{self.domain.replace(".","_")}.crt] file manually to [{directory}/server.crt] in order to proceed.', log_level=log_level)
		if not Files.exists(f"{directory}/server.crt") and Files.exists(f'{directory}/{self.domain.replace(".","_")}.crt'):
			os.system(f'mv {directory}/{self.domain.replace(".","_")}.crt {directory}/server.crt')
		if not Files.exists(f"{directory}/server.crt"):
			return Response.success(f'Failed to rename the [{directory}/{self.domain.replace(".","_")}.crt] file to [{directory}/server.crt].', log_level=log_level)

		# bundle ca.
		response = Code.execute(f"""
			cat {directory}/AAACertificateServices.crt {directory}/SectigoRSADomainValidationSecureServerCA.crt {directory}/Defaults.vars.userTrustRSAAAACA.crt > {self.database}/tls/server.ca-bundle
			cat {directory}/server.crt {self.database}/tls/server.ca-bundle > {self.database}/tls/signed.server.crt
			mv {self.database}/tls/server.crt {self.database}/tls/original.server.crt
			cp {self.database}/tls/signed.server.crt {self.database}/tls/server.crt
			""")
		if not response.success: return response
		if Files.exists(f"{self.database}/tls/signed.server.crt"):
			return Response.success(f"Successfully bundled ssl certificate [{directory}].", log_level=log_level)
		else:
			return Response.error(f"Failed to bundle ssl certificate [{directory}].", log_level=log_level)
	def check_dns(self, log_level=0):

		# loader.
		if log_level >= 0: loader = Console.Loader(f"Checking dns settings of domain {self.domain} ...")

		# check namecheap domain.
		response = self.namecheap.check_domain(self.namecheap.post_domain)
		if response.error != None: 
			if log_level >= 0: loader.stop(success=False)
			Response.log(response=response, log_level=log_level)
			return response
		elif not response["exists"]:
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Specified domain [{self.namecheap.post_domain}] is not owned by namecheap user [{self.namecheap.username}].", log_level=log_level)

		# add dns records.
		ip = NETWORK_INFO["public_ip"]
		if self.remote in ["vps"]: ip = self.vps_ip
		host = "@"
		www_host = "www"
		if self.namecheap.pre_domain != "":
			host = self.namecheap.pre_domain
			www_host = f"www.{self.namecheap.pre_domain}"
		response = self.namecheap.add_dns(
			# the domain.
			domain=self.namecheap.post_domain,
			# the dns record type,
			type="A",
			# the dns record host,
			host=www_host,
			# the dns record value/address,
			value=ip,)
		if response.error != None and "] already exists." not in response.error: 
			Response.log(response=response, log_level=log_level)
			if log_level >= 0: loader.stop(success=False)
			return response
		elif response.error == None: 
			Response.log(response=response, log_level=log_level)
		response = self.namecheap.add_dns(
			# the domain.
			domain=self.namecheap.post_domain,
			# the dns record type,
			type="A",
			# the dns record host,
			host=host,
			# the dns record value/address,
			value=ip,)
		if response.error != None and "] already exists." not in response.error: 
			Response.log(response=response, log_level=log_level)
			if log_level >= 0: loader.stop(success=False)
			return response
		elif response.error == None: 
			Response.log(response=response, log_level=log_level)
		
		# handlers.
		if log_level >= 0: loader.stop()
		return Response.success(f"Successfully checked the deployment dns settings for domain {self.domain}.", log_level=log_level)

		#




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils

# django imports.
from django.shortcuts import render
from django.http import HttpResponse, HttpResponseNotAllowed, JsonResponse
from django.urls import path, include
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator

# include apps from website/urls.py.
def include_apps(apps=[], auto_include=False):
	if auto_include:
		apps = []
		for _name_ in os.listdir(f"apps/"):
			if _name_ not in [".DS_Store"] and os.path.isdir(f"apps/{_name_}"): apps.append(_name_)
	urlpatterns = []
	for name in apps:
		_path_ = f"apps/{name}/"
		#try:
		if Files.exists(f"{_path_}/views.py"):
			l_path = f"{_path_}/views.py"
			urlpatterns.append(
				path('', include(f"apps.{name}.views"))
			)
		if Files.exists(f"{_path_}/requests.py"):
			l_path = f"{_path_}/requests.py"
			urlpatterns.append(
				path('', include(f"apps.{name}.requests"))
			)
		"""
		except Exception as e:
			sys.tracebacklimit = -1
			raise ImportError(f"Failed to import application [{name}], file [{l_path}], error: {e}.")
		"""
	return urlpatterns

# build url patterns.
def build_urls(views=[]):
	urls, landing_page = [], False
	for view in views:
		if view.type.lower() in ["view", "documentationview"] and not landing_page and view.landing_page:
			landing_page = True
			urls.append(path("", view.view))
			urls.append(path("/", view.view))
		urls.append(
			path(gfp.clean(view.url, remove_first_slash=True), view.view),
		)
	return urls
	
# the django request object class.
class Request(Object):
	def __init__(self,
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the requests id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# template data (optional).
		template_data={},
	):

		# defaultss.
		Object.__init__(self)
		self.parameters = Response.parameters

		# variables.
		self.base = base
		self.id = id
		if url == None:
			self.url = gfp.clean(f"{base}/{id}/", remove_first_slash=True)
		else:
			self.url = gfp.clean(url.replace("/$id",f"/{self.id}"), remove_last_slash=True, remove_first_slash=True)+"/"
		if self.base == None: 
			self.base = gfp.clean(FilePath(self.url).base(back=1), remove_last_slash=True)+"/"
		self.url = gfp.clean(self.url, remove_first_slash=True).replace(".html", "")
		self.base = gfp.clean(self.base, remove_first_slash=True)
		self.type = "request"
		self.template_data = template_data
	def view(self, request):
		# default view, user needs to create default def request(self, request)
		# can be overwritten with default def view(self, request):
		try: 
			response = self.request(request)
		except Exception as e: return self.response(utils.utils.catch_error(e))
		return self.response(response)
	def success(self, message, arguments={}):
		return Response.success(message, arguments, django=True)
	def error(self, error):
		return Response.error(error, django=True)
	def response(self, response):
		if isinstance(response, JsonResponse):
			return response
		else:
			try:
				return JsonResponse(response.dict(), safe=False)
			except AttributeError:
				return JsonResponse(response)
	def maintenance(self, request=None):
		return Response.error("Domain is under maintenance.")
	def permission_denied(self, request=None):
		return self.error("Permission denied.")
	# do not forget the self.parameters's functions.

# the django view object class.
class View(Object):
	def __init__(self, 
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the views id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# the html path (optional).
		html=None,
		# enable if this view is the [/] landing page.
		landing_page=False,
		# the template data (required).
		template_data={},
		# the object type (do not edit).
		type="View",
	):

		# defaults.
		Object.__init__(self)
		self.parameters = Response.parameters

		# vars.
		self.base = base
		self.id = id
		self.template_data = template_data
		self.landing_page = landing_page
		self.type = type
		if url == None:
			self.url = gfp.clean(f"{base}/{id}/", remove_first_slash=True)
		else:
			self.url = gfp.clean(url.replace("/$id",f"/{self.id}"), remove_last_slash=True, remove_first_slash=True)+"/"
		if self.base == None: 
			self.base = gfp.clean(FilePath(self.url).base(back=1), remove_last_slash=True)+"/"
		if html != None:
			if len(html) >= len(f"/apps/") and html[:len(f"/apps/")] == f"/apps/": html = html[len(f"/apps"):]
			elif len(html) >= len(f"apps/") and html[:len(f"apps/")] == f"apps/": html = html[len(f"apps/"):]
			self.html = gfp.clean(html.replace("/$id",f"/{self.id}"))
			self.app = gfp.clean(self.html, remove_first_slash=True, remove_last_slash=True, remove_double_slash=True).split("/")[0]
		else:
			self.app = gfp.clean(self.base, remove_first_slash=True, remove_last_slash=True, remove_double_slash=True).split("/")[0]
			l_base = gfp.clean(str(self.base), remove_last_slash=True)+"/"
			if len(l_base) >= len(f"{self.app}/{self.app}") and l_base[:len(f"{self.app}/{self.app}")] == f"{self.app}/{self.app}": l_base = l_base[len(f"{self.app}/"):]
			elif len(l_base) >= len(f"{self.app}/") and l_base[:len(f"{self.app}/")] == f"{self.app}/": l_base = l_base[len(f"{self.app}/"):]
			self.html = gfp.clean(f"{self.app}/html/{l_base}/{self.id}.html")
		self.base = gfp.clean(self.base, remove_last_slash=True)+"/"
		self.url = gfp.clean(self.url.replace(".html", ""), remove_first_slash=True, remove_last_slash=True)+"/"
		self.html = gfp.clean(self.html)
		self.type = type

		# check html.
		if self.type.lower() in ["view"] and "w3bsite/" not in self.html:
			path = FilePath(f"apps/{self.app}/")
			if Files.exists(path):
				path = Files.join(path, "html")
				if not Files.exists(path): 
					Response.log(f"&ORANGE&Creating&END& html base [{path}].")
					Files.create(path, directory=True)
				path = Files.join("apps", self.html)
				if not Files.exists(path):
					Response.log(f"&ORANGE&Creating&END& default html view [{path}].")
					base = FilePath(path).base(back=1)
					if not Files.exists(base): Files.create(base, directory=True)
					os.system(f"cp {SOURCE_PATH}/example/view.html {path}")

		#
	#@w3bsite.views.method_decorator(w3bsite.views.login_required)
	def view(self, request):
		return self.render(request)
	def render(self, request, 
		# overwrite default template data. #2
		template_data=None,
		# overwrite default html #3.
		html=None,
	):
		if template_data == None: template_data = self.template_data
		if html == None: html = self.html
		if isinstance(template_data, (Dictionary, Response.ResponseObject)): 
			template_data = template_data.raw()
		return render(request, html, template_data)
	def error(self, 
		# the django request parameter.
		request, 
		# the error title.
		title="Warning!",
		# the error title.
		message="Some error occured.",
		# the error icon (the static directory is root).
		icon="media/icons/warning.png", 
		# the redirect button text (right button).
		redirect_button="Ok",
		# the redirect url.
		redirect="/dashboard/home/",
		# overwrite default template data.
		template_data=None):
		if template_data == None: template_data = self.template_data
		title = title.replace("\n","")
		message = message.replace("\n","")
		template_data["ERROR"] = {
			"title":title,
			"message":message,
			"icon":icon,
			"redirect":redirect,
			"redirect_button":redirect_button,
		}
		if isinstance(template_data, (Dictionary, Response.ResponseObject)): 
			template_data = template_data.raw()
		return render(request, f"w3bsite/classes/apps/defaults/html/error.html", template_data)
	def maintenance(self, request, template_data=None):
		if template_data == None: template_data = self.template_data
		if isinstance(template_data, (Dictionary, Response.ResponseObject)): 
			template_data = template_data.raw()
		return render(request, f"w3bsite/classes/apps/defaults/html/maintenance.html", template_data)
	def permission_denied(self, request, template_data=None):
		if template_data == None: template_data = self.template_data
		if isinstance(template_data, (Dictionary, Response.ResponseObject)): 
			template_data = template_data.raw()
		return render(request, f"w3bsite/classes/apps/defaults/html/permission_denied.html", template_data)
	# do not forget the self.parameters's functions.









# template data object !!! DEPRICATED !!!.
class TemplateData(Object):
	def __init__(self, data={}):
		# defaults.
		Object.__init__(self)
		self.assign(data)
	# return raw data.
	def raw(self):
		return self.attributes()
		

#


#!/usr/bin/env python3

# imports.
from django.contrib import admin
from django.urls import path, include
from classes.config import *

# ______________________________________________________________________________________
#
# Inlude Apps.
#

# autmatically include all apps that contain a ./views.py or ./requests.py file.
imports = w3bsite.views.include_apps(auto_include=True)

# manually include apps.
#imports = w3bsite.views.include_apps([
#	"dashboard",
#)

# ______________________________________________________________________________________
#
# Authentication.
#

# import the default authentication requests.
imports += website.apps.authentication.requests.urls 

# import the default authentication views.
#imports += website.apps.authentication.views.urls 

# ______________________________________________________________________________________
#
# Payments.
#

# import the default payments requests.
#imports += website.apps.payments.requests.urls 

# import the default payments views.
#imports += website.apps.payments.views.urls 

# ______________________________________________________________________________________
#
# Defaults
#
urlpatterns = [
	# uncomment the following line to se the default django admin interface.
	#path('admin/', admin.site.urls)
] + imports



# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils
from w3bsite.classes import defaults as _defaults_

# the vps object class.
class VPS(_defaults_.Defaults):
	def __init__(self,
		# the remote ip address.
		ip=None,
		# the remote ssh port.
		port=22,
		# the remote username.
		username=None,
		# passed objects.
		namecheap=None,
		deployment=None,
		# defaults.
		defaults=None,

	):	

		# defaults.
		_defaults_.Defaults.__init__(self, traceback="w3bsite.Website.vps",)
		self.assign(defaults.dict())

		# check arguments.
		response = Response.parameters.check(
			traceback=self.__traceback__(),
			parameters={
				"ip":ip,
				"port":port,
				"username":username,
				"namecheap":namecheap,
				"deployment":deployment,
			})
		if not response.success: raise ValueError(response.error)
		
		# arguments.
		self.ip = ip
		self.port = port
		self.username = username

		# objects.
		self.deployment = deployment
		self.namecheap = namecheap

		# autocheck.
		#response = self.configure(log_level=-1)
		#if not response.success and "VPS is already live in production." not in response.error: raise ValueError(response.error)

		#
	def configure(self, reinstall=False, log_level=0):

		# check already live.
		#if self.live: return Response.error("VPS is already live in production.", log_level=log_level)

		# loader.
		if log_level >= 0: loader = Console.Loader(f"Configuring vps settings of website {self.domain} ...")

		# not live.
		if not self.live:

			# ssh key.
			if not Files.exists(f"{self.root}/__defaults__/ssh"): 
				response = ssht00ls.keys.generate(
					path=f"{self.root}/__defaults__/ssh", 
					comment=f"SSH key for domain {self.domain}", 
					passphrase="",)
				if not response.success: 
					if log_level >= 0: loader.stop(success=False)
					return response

			# ssh config.
			response = ssht00ls.aliases.check(
				alias=self.domain, 
				create=True,
				info={
					"username":self.username, 
					"public_ip":self.ip,
					"private_ip":self.ip,
					"public_port":self.port,
					"private_port":self.port,
					"private_key":f"{self.root}/__defaults__/ssh/private_key",
					"public_key":f"{self.root}/__defaults__/ssh/public_key",
					"passphrase":"",
				})
			if not response.success: 
				if log_level >= 0: loader.stop(success=False)
				return response

		# handler.
		if log_level >= 0: loader.stop()
		return Response.success("Successfully configured the vps settings.", log_level=log_level)

		#
	def deploy(self, code_update=False, reinstall=False, log_level=0):

		# configure vps.
		response = self.configure(reinstall=reinstall, log_level=log_level)
		if not response.success: return response

		# configure deployment.
		response = self.deployment.configure(reinstall=reinstall, log_level=log_level)
		if not response.success: return response

		# loader.
		if log_level >= 0: loader = Console.Loader(f"Configuring deployment of website {self.domain} to vps {self.ip} ...")

		# check already live.
		if self.live: 
			if log_level >= 0: loader.stop(success=False)
			return Response.error("VPS is already live in production.", log_level=log_level)

		# check installer & requirements file.
		if not Files.exists(f"{self.root}/requirements/installer"):
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Required requirements file {self.root}/requirements/installer does not exist.", log_level=log_level)
		if not Files.exists(f"{self.root}/requirements/requirements.pip"):
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Required requirements file {self.root}/requirements/requirements.pip does not exist.", log_level=log_level)
		data = Files.load(f"{self.root}/requirements/installer")
		installed_alias = data.split('alias="')[1].split('"')[0]
		#installed_location = data.split('package="')[1].split('"')[0].replace("$alias", installed_alias)
		installed_location = self.library

		# check connection.
		if log_level >= 0: loader.mark(new_message=f"Attempting to connect with vps {self.ip}")
		output = dev0s.utils.__execute_script__(f"printf 'yes\n' | ssh {self.domain} ' echo Hello World ' ").replace('\n\n','\n')
		if "Permission denied (publickey)" in output:
			if log_level >= 0:  loader.stop(success=False)
			data = Files.load(f'{self.root}/__defaults__/ssh/public_key').replace('\n','')
			return Response.error(f"Unable to connect with {self.domain} over ssh (permission denied). Did you install public key {data} into the vps server?", log_level=log_level)
		elif "Hello World" not in output:
			if log_level >= 0:  loader.stop(success=False)
			return Response.error(f"Unable to connect with {self.domain} over ssh, ssh output: \n{output}.", log_level=log_level)

		# installer arguments.
		installer_arguments = ""
		if code_update: installer_arguments += " --code-update"
		if reinstall: installer_arguments += " --reinstall"

		# copy current source to vps & installer script.		
		if log_level >= 0: loader.mark(new_message=f"Installing source code of website {self.domain} on vps {self.ip}")
		package_name = gfp.name(path=self.root)
		if code_update:
			tmp = self.library
		else:
			tmp = f"/tmp/{package_name}/"
		response = ssht00ls.ssync.push(
			alias=self.domain,
			path=self.root,
			remote=tmp,
			checks=False,
			directory=True,
			delete=True,
			check_base=False,)
		if not response.success: 
			if log_level >= 0: loader.stop(success=False)
			return Response.error(f"Failed to deploy website {self.domain} on vps {self.ip}, error (#1): {response.error}.", log_level=log_level)
		#Response.log(f"&ORANGE&Do not forget&END& to remove the [{tmp}/website.py] file.")
		#response = ssht00ls.ssh.command(
		#	alias=self.domain,
		#	command=f"rm -fr {tmp}/website.py",
		#	log_level=-1,)
		#if not response.success: return response

		# execute installer script.
		if not code_update:
			if log_level >= 0: loader.mark(new_message=f"Executing installer script of website {self.domain} on vps {self.ip}")
			response = ssht00ls.ssh.command(
				alias=self.domain,
				command=f"chmod +x {tmp}/requirements/installer && bash {tmp}/requirements/installer{installer_arguments}",
				log_level=-1,)
			if not response.success: 
				if log_level >= 0: loader.stop(success=False)
				return Response.error(f"Failed to deploy website {self.domain} on vps {self.ip}, error (#2): {response.error}.", log_level=log_level)

			# deploy.
			if log_level >= 0: loader.mark(new_message=f"Deploying domain {self.domain} on vps {self.ip}")
			response = ssht00ls.ssh.command(
				alias=self.domain,
				command=f"python3 {installed_location}/website.py --deploy{installer_arguments} --non-interactive",
				log_level=-1,)
			if not response.success:
				if log_level >= 0: loader.stop(success=False)
				return Response.error(response.error, log_level=log_level)
			elif "Successfully deployed website https://" not in response.output:
				if log_level >= 0: loader.stop(success=False)
				return Response.error(f"Failed to deploy website {self.domain} on vps {self.ip}, error (#3): {response.output}.", log_level=log_level)

		# code update restart.
		else:
			response = ssht00ls.ssh.command(
				alias=self.domain,
				command=f"python3 {self.library}/website.py --restart --non-interactive",
				log_level=-1,)
			if not response.success: return response

		# handler.
		if log_level >= 0: loader.stop()
		return Response.success(f"Successfully deployed website {self.domain} on vps {self.ip}.", log_level=log_level)

		#



# imports.
from w3bsite.classes.config import *
import zlib, base64, binascii, glob
from Crypto.PublicKey import RSA as _RSA_
from Crypto.Cipher import PKCS1_OAEP

# the rsa object class.
class RSA(object):
	def __init__(self, 
		# the private key.
		private_key=None,
		# the public key.
		public_key=None,
	):
		self.public_key_pem = public_key
		self.private_key_pem = private_key
		self.public_key = None
		self.private_key = None
	# key management.
	def generate_keys(self):
			
		# generate.
		key_pair = _RSA_.generate(4096, e=65537)
		public_key = key_pair.publickey()
		public_key_pem = public_key.exportKey()
		private_key_pem = key_pair.exportKey()


		# save.
		self.private_key_pem = private_key_pem.decode()
		self.public_key_pem = public_key_pem.decode()
		
		# response.
		return Response.success("Successfully generated a key pair.")

		#
	def load_private_key(self):

		# load keys.
		private_key = self.private_key_pem
		if private_key == None:
			return Response.error("The private key is undefined.")

		# initialize keys.
		try:
			private_key = _RSA_.importKey(private_key)
		except Exception as e:
			return Response.error(f"Exception: {e}")
		
		# response.
		self.private_key = private_key
		return Response.success("Successfully loaded the private key.")

		#
	def load_public_key(self):

		# load keys.
		public_key = self.public_key_pem
		if public_key == None:
			return Response.error("The public key is undefined.")

		# initialize keys.
		try:
			public_key = _RSA_.importKey(public_key)
		except Exception as e:
			return Response.error(f"Exception: {e}")
		
		# response.
		self.public_key = public_key
		return Response.success("Successfully loaded the public key.")

		#
	# encrypting & decrypting.
	def encrypt(self, string):

		# encrypt.
		encrypted = self.__encrypt_blob__(string, self.public_key)

		# response.
		return Response.success(f"Successfully encrypted the string.", {
			"encrypted": encrypted,
		})

		#
	def decrypt(self, string):
		
		# decrypt.
		if isinstance(string,str): string = string.encode()
		decrypted = self.__decrypt_blob__(string, self.private_key)
		
		# response.
		return Response.success(f"Successfully decrypted the string.", {
			"decrypted": decrypted,
		})

		#
	# system functions.
	def __encrypt_blob__(self, blob, public_key, silent=True):
		#Import the Public Key and use for encryption using PKCS1_OAEP
		rsa_key = public_key
		rsa_key = PKCS1_OAEP.new(rsa_key)

		#compress the data first
		try: 
			blob = zlib.compress(blob.encode())
		except: 
			blob = zlib.compress(blob)
		
		#In determining the chunk size, determine the private key length used in bytes
		#and subtract 42 bytes (when using PKCS1_OAEP). The data will be in encrypted
		#in chunks
		chunk_size = 470
		offset = 0
		end_loop = False
		encrypted = bytearray()
		max_offset, progress = len(blob), 0
		if silent == False: print(f'Encrypting {max_offset} bytes.')
		while not end_loop:
			#The chunk
			chunk = blob[offset:offset + chunk_size]

			#If the data chunk is less then the chunk size, then we need to add
			#padding with " ". This indicates the we reached the end of the file
			#so we end loop here
			if len(chunk) % chunk_size != 0:
				end_loop = True
				#chunk += b" " * (chunk_size - len(chunk))
				chunk += bytes(chunk_size - len(chunk))
			#Append the encrypted chunk to the overall encrypted file
			encrypted += rsa_key.encrypt(chunk)

			#Increase the offset by chunk size
			offset += chunk_size
			l_progress = round((offset/max_offset)*100,2)
			if l_progress != progress:
				progress = l_progress
				if silent == False: print('Progress: '+str(progress), end='\r')

		#Base 64 encode the encrypted file
		return base64.b64encode(encrypted)
	def __decrypt_blob__(self, encrypted_blob, private_key, silent=True):

		#Import the Private Key and use for decryption using PKCS1_OAEP
		rsakey = private_key
		rsakey = PKCS1_OAEP.new(rsakey)

		#Base 64 decode the data
		encrypted_blob = base64.b64decode(encrypted_blob)

		#In determining the chunk size, determine the private key length used in bytes.
		#The data will be in decrypted in chunks
		chunk_size = 512
		offset = 0
		decrypted = bytearray()
		max_offset = len(encrypted_blob)
		progress = 0
		if silent == False: print(f'Decrypting {max_offset} bytes.')
		#keep loop going as long as we have chunks to decrypt
		while offset < len(encrypted_blob):
			#The chunk
			chunk = encrypted_blob[offset: offset + chunk_size]

			#Append the decrypted chunk to the overall decrypted file
			decrypted += rsakey.decrypt(chunk)

			#Increase the offset by chunk size
			offset += chunk_size
			l_progress = round((offset/max_offset)*100,2)
			if l_progress != progress:
				progress = l_progress
				if silent == False: print('Progress: '+str(progress), end='\r')

		#return the decompressed decrypted data
		return zlib.decompress(decrypted)



# imports
from w3bsite.classes.views.defaults import *

# the documentations object class.
import importlib
class Documentations(View):
	def __init__(self):
		a=1
	def chapter(self,
		# the chapter title.
		title="Chapter Title",
		# the chapter's sections.
		sections={},
		# whether the chapter is folded in the leftbar or not.
		folded=True,
	):
		return {
			"title":title,
			"sections":self.__assign_element_ids__(sections, key="element_id"),
			"sections_count":len(sections),
			"folded":folded,
			"element_id":self.__generate_element_id__(),
		}
	def section(self,
		title="Section Title",
		description=None,
		attributes={},
		parameters={},
		code_areas={},
	):

		description = self.__color_description__(description)
		return {
			"title":title,
			"description":description,
			"description_count":len(description),
			"attributes":self.__assign_element_ids__(attributes, key="element_id"),
			"attributes_count":len(attributes),
			"parameters":self.__assign_element_ids__(parameters, key="element_id"),
			"parameters_count":len(parameters),
			"code_areas":self.__assign_element_ids__(code_areas, key="element_id"),
			"code_areas_count":len(code_areas),
			"element_id":self.__generate_element_id__(),
		}
	def parameter(self,
		# the parameters id.
		id="my_variable",
		# the parameters format (string / list with strings).
		format=None,
		# the parameters description.
		description=None,
		# whether the parameter is required or optional.
		required=False,
		# whether the parameter is folded by default or not.
		folded=False,
	):
		return self.__parameter_attribute__(
			id=id,
			format=format,
			description=self.__color_description__(description),
			required=required,
			folded=folded,)
	def attribute(self,
		# the attributes id.
		id="my_variable",
		# the attributes format (string / list with strings).
		format=None,
		# the attributes description.
		description=None,
		# whether the attribute is folded by default or not.
		folded=False,
	):
		return self.__parameter_attribute__(
			id=id,
			format=format,
			description=self.__color_description__(description),
			folded=folded,)
	def code_area(self,
		# the header title.
		title="my_variable",
		# the header subtitle.
		substitle=None,
		# the code area's content.
		content={
			"json":{},
			"python":"",
			# use your own. Unkown types are just non colored strings.
		},
		# the content css style with python_like or jsLike keys.
		styles={
			"height":"auto",
		},
	):
		# colors.
		def process_data(_data_):

			# normalize.
			_data_ = _data_.replace("	", "    ")
			data = _data_
			for key, replacer in [
				["\ndef  ", "\ndef "],
				["\nclass  ", "\nclass "],
			]:
				while True:
					if key in data: data = data.replace(key, replacer)
					else: break

			# chunk into array with
			parentheses = 0
			items = [] # [ [ $type, $lines ] ]
			chars, previous, lastchar, string_recognizer, string_open, string_closed = "", "default", None, None, False, False
			linecount = 0
			for char in data:
				if char == "\n":
					linecount += 1
				add_char = True
				refresh_parentheses = False
				if char == "(":
					if parentheses > 0:
						refresh_parentheses = True
					parentheses += 1
				elif char == ")":
					parentheses -= 1
					if parentheses > 0:
						refresh_parentheses = True
				if previous not in ["comment", "string"] and refresh_parentheses:
					items.append([previous, chars+char])
					chars = ""
					previous = "parentheses"
					add_char = False
				elif char == "#" and previous not in ["comment"]:
					items.append([previous, chars])
					chars = ""
					previous = "comment"
				elif previous == "comment" and char == "\n":
					items.append([previous, chars+char])
					add_char = False
					chars = ""
					previous = "default"
				elif previous not in ["comment"] and char in ["'", '"', "`"]:
					if previous in ["parentheses", "default"]:
						items.append([previous, chars])
						chars = ""
						previous = "string"
						string_recognizer = char
					elif previous == "string":
						if string_recognizer == char and lastchar != char:
							string_recognizer = None
							add_char = False
							items.append(["string", chars+char])
							chars = ""
							previous = "default"
					elif previous != None: raise ValueError(f"Unkown previous: {previous}")
				elif previous not in ["comment", "string"] and parentheses > 0:
					#if previous in ["string", "default", "comment"]:
					if previous != "parentheses":
						add_char = False
						items.append([previous, chars+char])
						chars = ""
						previous = "parentheses"
				elif previous == "parentheses" and parentheses <= 0:
					items.append([previous, chars])
					chars = ""
					previous = "default"
				elif previous not in ["string", "comment"]:
					if previous in ["string", "comment", "parentheses"]:
						items.append([previous, chars])
						chars = ""
					elif previous not in ["default",None]: raise ValueError(f"Unkown previous: {previous}")
					previous = "default"
				lastchar = char
				if add_char: chars += char

			# set remainder.
			items.append([previous, chars])
			return items

			#

		# color code.
		default = None
		words = {}
		for language, data in content.items():
			default = language
			if language == "json":

				# language functions.
				def process_items(items):

					# iterate items.
					words, count, max, next, next_chars, previous, previous_chars = [], 0, len(items)-1, None, None, None, None
					for type, chars in items:
						#print(type.upper(), chars)

						# get previous.
						if count == 0: previous = None
						else: previous,previous_chars = items[count-1]
						
						# get next.
						if count == max: next = None
						else: next,next_chars = items[count+1]

						# string.
						if type == "string":
							words.append(self.__create_word__(word=chars, color="yellow", type="string", joiner="", language=language))

						# comment.
						elif type == "comment":
							words.append(self.__create_word__(word=chars, color="grey", type="comment", joiner="", language=language))

						# else
						elif type == "default":
							words += process_default(chars, next=next, next_chars=next_chars)

						# increment count.
						count += 1

					# return.
					return words
				def process_default(chars, next=None, next_chars=None,):
					words = []
					last, keep = {}, 100
					for i in range(keep+1):
						if i > 0: last[str(i)] = ""
					_chars_, int_chars = "", ""
					integer_block = False
					for char in chars:
						add_char = True

						# integer chars.
						try: float(int_chars+char) ; integer = True
						except: integer = False
						if integer:
							integer_block = True
							int_chars += char
						elif int_chars != "":
							integer_block = False
							words.append(self.__create_word__(word=_chars_[:-len(int_chars)], joiner="", language=language))
							words.append(self.__create_word__(word=int_chars, color="purple", joiner="", language=language))
							int_chars = ""
							_chars_ = ""

						# add char.
						if add_char:
							_chars_ += char

						# last.
						for i in range(keep+1):
							if i > 0:
								if len(last[str(i)]) >= i:
									last[str(i)] = last[str(i)][1:]
								last[str(i)] += char

						# no integer.
						if not integer_block:

							# purple.
							for i in [
								" false ", " true ", " null ",
								" false,", " true,", " null,",
								":false ", ":true ", ":null ",
								":false,", ":true,", ":null,",
								":false\n", ":true\n", ":null\n",
								" false\n", " true\n", " null\n",
							]:
								if last[str(len(i))] == i:
									words.append(self.__create_word__(word=_chars_[:-len(i)], joiner="", language=language))
									words.append(self.__create_word__(word=i, color="purple", joiner="", language=language))
									_chars_ = ""
				
					# default remainder.
					words.append(self.__create_word__(word=_chars_, joiner="", language=language))
					return words

					#


				# process.
				if isinstance(data, dict):
					data = json.dumps(data, indent=4)
				elif not isinstance(data, str): raise ValueError(f"Unkown json code area content format: {data}")
				items = process_data(data)
				words[language] = process_items(items)

				#
			elif language == "python":

				# language functions.
				def process_items(items):

					# iterate items.
					words, count, max, next, next_chars, previous, previous_chars = [], 0, len(items)-1, None, None, None, None
					for type, chars in items:
						#print(type.upper(), chars)

						# get previous.
						if count == 0: previous = None
						else: previous,previous_chars = items[count-1]
						
						# get next.
						if count == max: next = None
						else: next,next_chars = items[count+1]

						# string.
						if type == "string":
							words.append(self.__create_word__(word=chars, color="yellow", type="string", joiner="", language=language))

						# comment.
						elif type == "comment":
							words.append(self.__create_word__(word=chars, color="grey", type="comment", joiner="", language=language))

						# parentheses.
						elif type == "parentheses":
							words += process_parentheses(chars, previous=previous, previous_chars=previous_chars)

						# default.
						elif type == "default":
							words += process_default(chars, next=next, next_chars=next_chars)

						# increment count.
						count += 1

					# return.
					return words
				def process_parentheses(chars, previous=None, previous_chars=None):
					
					# words.
					words = []
					
					# by class initialization.
					if previous_chars != None  and "class " in previous_chars:
						c, _chars_, last, spaces, success = len(previous_chars)-1, "", "", 0, False
						for _ in previous_chars:
							char = previous_chars[c]
							if char == " ": 
								if spaces > 0:
									success = False
									break 
								spaces += 1
							elif last == " ssalc":
								success = True
								break
							_chars_ += char
							if len(last) >= 6:
								last = last[1:]
							last += char
							c -=1
						if last == " ssalc" or success:
							for item in chars.split(","):
								words.append(self.__create_word__(word=item, color="green", italic=True, joiner=",", language=language))
							return words
					
					# by no class initialization.
					items = []
					_chars_, comma = "", True
					for char in chars:
						add_char = True
						if char == ",":
							comma = True
							if _chars_ != "":
								items.append(_chars_)
							_chars_ = ""
						elif comma:
							comma = False
							items.append(_chars_)
							_chars_ = ""
						if add_char: _chars_ += char
					items.append(_chars_)
					#
					for item in items:
						if item not in [""]:
							if item in [",", "\n"]:
								words.append(self.__create_word__(word=item, color=None, joiner="", language=language))
							elif "=" not in item:
								if "(" in item:
									slicer_id = utils.__first_occurence_reversed__(string=item, charset=[" ", "\n", "\r"])
									if slicer_id == None: slicer_id = " "
									before, after = utils.__before_after_last_occurence__(string=item, slicer=slicer_id, include_before=True)
									if len(after) > 0:
										words.append(self.__create_word__(word=before, joiner="", language=language))
									else:
										after = before
									if "." not in after:
										words.append(self.__create_word__(word=after[:-1], color="blue", joiner="", language=language))
										words.append(self.__create_word__(word="(", joiner="", language=language))
									else:
										before, after = utils.__before_after_last_occurence__(string=after, slicer=".", include_before=True)
										words.append(self.__create_word__(word=before, joiner="", language=language))
										words.append(self.__create_word__(word=after[:-1], color="blue", joiner="", language=language))
										words.append(self.__create_word__(word="(", joiner="", language=language))
								elif item == "...":
									words.append(self.__create_word__(word=item, color="orange", italic=True, joiner="", language=language))
								else:
									words.append(self.__create_word__(word=item, color="orange", italic=True, joiner="", language=language))
							else:
								before, after = utils.__before_after_first_occurence__(string=item, slicer="=", include=False)
								words.append(self.__create_word__(word=before, color="orange", italic=True, joiner="", language=language))
								words.append(self.__create_word__(word="=", color="red", joiner="", language=language))
								if "(" in after:
									if "." not in after:
										words.append(self.__create_word__(word=after[:-1], color="blue", joiner="", language=language))
										words.append(self.__create_word__(word="(", joiner="", language=language))
									else:
										before, after = utils.__before_after_last_occurence__(string=after, slicer=".", include_before=True)
										words.append(self.__create_word__(word=before, joiner="", language=language))
										words.append(self.__create_word__(word=after[:-1], color="blue", joiner="", language=language))
										words.append(self.__create_word__(word="(", joiner="", language=language))
								else:
									afters, _chars_ = [], ""
									c, max = 0, len(after)-1
									for char in after:
										add_char = True
										try: float(_chars_) ; integer = True
										except: integer = False
										if integer:
											try: float(_chars_+char) ; new_integer = True
											except: new_integer = False
											if new_integer:
												a=1
											else:
												afters.append(self.__create_word__(word=_chars_, color="purple", joiner="", language=language))
												_chars_ = ""
										elif _chars_ in ["None", "True", "False"]:
											afters.append(self.__create_word__(word=_chars_, color="purple", joiner="", language=language))
											_chars_ = ""
										if add_char: _chars_ += char
									if _chars_ != "":
										try: float(_chars_) ; integer = True
										except: integer = False
										if integer:
											afters.append(self.__create_word__(word=_chars_, color="purple", joiner="", language=language))
										elif _chars_ in ["None", "True", "False"]:
											afters.append(self.__create_word__(word=_chars_, color="purple", joiner="", language=language))
										else:
											afters.append(self.__create_word__(word=_chars_, color=None, joiner="", language=language))
									c = 0
									for after in afters:
										if c == len(afters)-1:
											after["joiner"] = ""
											words.append(after)
										else:
											words.append(after)
										c += 1
					return words
				def process_default(chars, next=None, next_chars=None,):
					words = []
					parentheses_end = chars[len(chars)-1] == "("
					last, keep = {}, 100
					for i in range(keep+1):
						if i > 0: last[str(i)] = ""
					_chars_, slices, slicer_count, catch_function = "", 0, 0, False
					integer_block, int_chars, slicer_id = False, "", None
					if parentheses_end:
						slicer_id = utils.__first_occurence_reversed__(string=chars, charset=[" ", "\n", "\r"]) 
						slicer_count = chars.count(slicer_id)
					firstline, lastchar, nextchar, charcount = True, None, None, 0
					for char in chars:
						if char in ["\n"]: firstline = False
						try: nextchar = chars[charcount+1]
						except: nextchar = None
						add_char = True

						# classes & functions.
						if char == slicer_id:
							slices += 1
						if parentheses_end and slicer_count > 0 and slices == slicer_count:
							if not catch_function:
								if last["5"] == "class":
									words.append(self.__create_word__(word=_chars_[:-len("class")], joiner="", language=language))
									words.append(self.__create_word__(word="class", color="blue", joiner="", language=language))
								elif last["3"] == "def":
									words.append(self.__create_word__(word=_chars_[:-len("def")], joiner="", language=language))
									words.append(self.__create_word__(word="def", color="blue", joiner="", language=language))
								elif last[str(len("__init__"))] == "__init__":
									words.append(self.__create_word__(word=_chars_[:-len("__init__")], joiner="", language=language))
									words.append(self.__create_word__(word="__init__", color="blue", joiner="", language=language))
								else:
									words.append(self.__create_word__(word=_chars_, joiner=""))
								_chars_ = ""
							elif char == "(":
								if last["5"] == "class":
									words.append(self.__create_word__(word=_chars_, color="green", joiner="", language=language))
								elif last["3"] == "def":
									words.append(self.__create_word__(word=_chars_, color="green", joiner="", language=language))
								else:
									if "." in _chars_:
										before, after_1 = utils.__before_after_last_occurence__(string=_chars_, slicer=".", include_before=True)
										first = utils.__first_occurence_reversed__(string=before, charset=[" string.", "\nstring."])
										if first != None:
											before, selected, after_2 = utils.__before_selected_after_first_occurence__(string=before, slicer="first")
											words.append(self.__create_word__(word=before, joiner="", language=language))
											words.append(self.__create_word__(word=selected, color="orange", italic=True, joiner="", language=language))
											words.append(self.__create_word__(word=after_2, color="blue", joiner="", language=language))
										else:
											words.append(self.__create_word__(word=before, joiner="", language=language))
										words.append(self.__create_word__(word=after_1, color="blue", joiner="", language=language))
									else:
										words.append(self.__create_word__(word=_chars_, color="blue", joiner="", language=language))
								_chars_ = ""
							catch_function = True
							_chars_ += char

						# else.
						else:

							# integer chars.
							try: 
								float(int_chars+char)
								integer = True
							except: integer = False
							if integer:
								integer_block = True
								int_chars += char
							elif int_chars != "":
								integer_block = False
								"""
								nointeger = False
								for a in "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM":
									for b in "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM":
										l_lastchar , l_nextchar = lastchar, nextchar
										if l_lastchar == None: l_lastchar = ""
										if l_nextchar == None: l_nextchar = ""
										if len(f"{l_nextchar}{int_chars}{l_nextchar}") < len(f"{a}{int_chars}{b}"): 
											nointeger = None
											break 
										if f"{a}{int_chars}{b}" == f"{l_nextchar}{int_chars}{l_nextchar}":
											nointeger = True
											break
									if nointeger in [True, None]: break
								"""
								if last[str(len(i)+1)] not in [f" {int_chars}",f"\n{int_chars}",f"\r{int_chars}",f"({int_chars}",f"={int_chars}",f"+{int_chars}",f"-{int_chars}","{"+f"{int_chars}"]:
								#if lastchar not in [" ","\n","\r","(","=","+","-","{"]:
									int_chars = ""
								else:
									words.append(self.__create_word__(word=_chars_[:-len(int_chars)], joiner="", language=language))
									words.append(self.__create_word__(word=int_chars, color="purple", joiner="", language=language))
									int_chars = ""
									_chars_ = ""
							"""
							try: float(_chars_) ; integer = True
							except: integer = False
							if integer:
								try: float(_chars_+char) ; new_integer = True
								except: new_integer = False
								if char == "." or new_integer:
									integer_block = True
								else:
									words.append(self.__create_word__(word=_chars_, color="purple", joiner="", language=language))
									_chars_ = ""
									integer_block = False
							else: integer_block = False
							"""

							# add char.
							if add_char:
								_chars_ += char

							# last.
							for i in range(keep+1):
								if i > 0:
									if len(last[str(i)]) >= i:
										last[str(i)] = last[str(i)][1:]
									last[str(i)] += char

							# no integer.
							if not integer_block:

								# red.
								local = []
								for i in ["if", "for", "elif", "else", "elif", "raise", "break", "try", "except", "return", "import", "del", "print"]:
									if firstline or last[str(len(i)+1)] == last[str(len(i))]:
										local.append(f"{i} ")
										local.append(f"{i}\n")
										local.append(f"{i}\r")
									local.append(f" {i} ")
									local.append(f" {i}\n")
									local.append(f" {i}\r")
									local.append(f"\n{i} ")
									local.append(f"\n{i}\n")
									local.append(f"\n{i}\r")
									local.append(f"\r{i} ")
									local.append(f"\r{i}\n")
									local.append(f"\r{i}\r")
								for i in [
									"+", "|", "&", "-", "=", ">", "<", "!", "@", 
									" in ", " or ",
									" try:","\ntry:"," except:","\nexcept:"," else:","\nelse:",
								]+local:
									#print(f"[{last[str(len(i))]}] VS [{i}], lastchar: [{lastchar}]")
									if last[str(len(i))] == i:
										words.append(self.__create_word__(word=_chars_[:-len(i)], joiner="", language=language))
										words.append(self.__create_word__(word=i, color="red", joiner="", language=language))
										_chars_ = ""

								# orange italic.
								local = []
								for i in ["self.", "self[", ]:
									if firstline or last[str(len(i)+1)] == last[str(len(i))]:
										local.append(f"{i}")
									local.append(f" {i}")
									local.append(f"\n{i}")
									local.append(f"\r{i}")
								for i in [
								]+local:
									if last[str(len(i))] == i:
										words.append(self.__create_word__(word=_chars_[:-len(i)], joiner="", language=language))
										words.append(self.__create_word__(word=i, color="orange", italic=True, joiner="", language=language))
										_chars_ = ""
								
								# purple.
								for i in [
									" False ", " True ", " None ",
									" False\n", " True\n", " None\n",
									" False\r", " True\r", " None\r",
									"...",
								]:
									#if i == "...":
									#	print(f"{i} VS [{last[str(len(i))]}]")
									if last[str(len(i))] == i:
										words.append(self.__create_word__(word=_chars_[:-len(i)], joiner="", language=language))
										words.append(self.__create_word__(word=i, color="purple", joiner="", language=language))
										_chars_ = ""
								
								# blue.
								for i in [
								]:
									if last[str(len(i))] == i:
										words.append(self.__create_word__(word=_chars_[:-len(i)], joiner="", language=language))
										words.append(self.__create_word__(word=i, color="blue", joiner="", language=language))
										_chars_ = ""

						# last char.
						lastchar = char
						charcount += 1

				
					# default remainder.
					words.append(self.__create_word__(word=_chars_, joiner="", language=language))
					return words

					#

				# process.
				if not isinstance(data, str):
					#lines = str(content[language]).split("\n")
					raise ValueError(f"Content of language {language} is not str format, {data}.")
				items = process_data(data)
				words[language] = [self.__create_word__(word="\n")] + process_items(items)

				#
			elif language in ["bash", "shell", "cli", "pip", "pip3", "curl"]:

				# language functions.
				def process_items(items):

					# iterate items.
					words, count, max, next, next_chars, previous, previous_chars = [], 0, len(items)-1, None, None, None, None
					for type, chars in items:
						#print(type.upper(), chars)

						# get previous.
						if count == 0: previous = None
						else: previous,previous_chars = items[count-1]
						
						# get next.
						if count == max: next = None
						else: next,next_chars = items[count+1]

						# string.
						if type == "string":
							words.append(self.__create_word__(word=chars, color="yellow", type="string", joiner="", language=language))

						# comment.
						elif type == "comment":
							words.append(self.__create_word__(word=chars, color="grey", type="comment", joiner="", language=language))

						# else
						elif type == "default":
							words += process_default(chars, next=next, next_chars=next_chars)

						# increment count.
						count += 1

					# return.
					return words
				def process_default(chars, next=None, next_chars=None,):
					words = []
					delimiters = [" ","\n"]
					scentence = ""
					word = ""
					for char in chars:

						# check delimiter.
						if char in delimiters:
							if (len(word) > 0 and word[0] == "-") or (len(word) > 1 and word[:2] == " -"):
								words.append(self.__create_word__(word=word, color="orange", italic=True, joiner="", language=language))
							elif len(word) > 0 and scentence == "":
								words.append(self.__create_word__(word=word, color="blue", joiner="", language=language))
							else:
								words.append(self.__create_word__(word=word, joiner="", language=language))
							scentence += word
							word = ""
							if char in "\n": scentence = ""


						# add word.
						word += char

				
					# default remainder.
					if word != "":
						if (len(word) > 0 and word[0] == "-") or (len(word) > 1 and word[:2] == " -"):
							words.append(self.__create_word__(word=word, color="orange", italic=True, joiner="", language=language))
						elif len(word) > 0 and scentence == "":
							words.append(self.__create_word__(word=word, color="blue", joiner="", language=language))
						else:
							words.append(self.__create_word__(word=word, joiner="", language=language))
					return words

					#


				# process.
				if isinstance(data, dict):
					data = json.dumps(data, indent=4)
				elif not isinstance(data, str): raise ValueError(f"Unkown json code area content format: {data}")
				items = process_data(data)
				words[language] = process_items(items)

			else:

				# no color.
				words[language] = [self.__create_word__(word=data, language=language)]

		# return.
		return {
			"title":title.upper(),
			"substitle":substitle,
			"words":words,
			"styles":styles,
			"default":default,
			"element_id":self.__generate_element_id__(),
		}
	# system functions.
	def __parameter_attribute__(self,
		id="my_variable",
		format=None,
		description=None,
		required=False,
		folded=None,
	):
		if isinstance(format, str):
			format = [format]
		elif not isinstance(format, list):
			raise ValueError(f"Unkown format instance [{format}]. The format should be a list with formats or a single string format.")
		return {
			"id":id,
			"format":format,
			"description":description,
			"description_count":len(description),
			"folded":folded,
			"required":required,
			"element_id":self.__generate_element_id__(),
		}
	def __generate_element_id__(self, joiner=""):
		return joiner+String("").generate(length=32, digits=True, capitalize=True)
	def __create_word__(self, color=None, italic=False, type=None, word="", linecount=None, joiner="", language=None, styles={}):
		colors = {
			"red":"#FF0071",
			"green":"#93E600",
			"blue":"#25DBF2",
			"orange":"#FF9100",
			"yellow":"#E9DC61",
			"purple":"#B57AFF",
			"grey":"#75705B",
		}
		try: color = colors[color]
		except: color = None
		return {
			"value":word,
			"color":color,
			"italic":italic,
			"linecount":linecount,
			"language":language,
			"joiner":joiner,
			"type":type,
			"styles":styles,
		}
	def __color_description__(self, description):
		def clean_tag(tag):
			while True:
				if " " in tag: tag = tag.replace(" ", "")
				else: break
			return tag
		words, c, open, closing, chars, max = [], 0, False, False, "", len(description)
		while c < max:
			add_char = True
			char = description[c]
			try: next = description[c+1]
			except: next = None
			if open and (char == "<" or (closing and char == ">")):
				if char == "<":
					closing = True
				elif closing and char == ">":
					add_char = False
					chars += char
					words.append(self.__create_word__(word=chars))
					chars = ""
					closing = False
					open = False
			elif not open and char == "<":
				open = True
				words.append(self.__create_word__(word=chars))
				chars = ""
			if add_char: chars += char
			c += 1
		if chars != "":
			words.append(self.__create_word__(word=chars))
		_words_ = []
		for word in words:
			if len(word["value"]) > 0 and word["value"][0] == "<":
				# style tags.
				tag = clean_tag(word["value"].split(">")[0]+">")
				word["value"] = word["value"][len(tag):]
				word["value"], _ = utils.__before_after_last_occurence__(string=word["value"], slicer="<", include=False)
				styles = {}
				# append styles.
				block = False
				if not block and tag in ["<bold>"]:
					block = True
					styles = utils.__append_dict__(old=styles, overwrite=True, new={
						"font_weight":"700",})
				if not block and tag in ["<italic>"]:
					block = True
					styles = utils.__append_dict__(old=styles, overwrite=True, new={
						"font_style":"italic",})
				if not block and tag in ["<code>", "<parameter>", "<attribute>"]:
					styles = utils.__append_dict__(old=styles, overwrite=True, new={
						"background": f"#00000025", 
						"border":"1px solid #00000035",
						"border_radius":"10px",
						"font_family":"Courier New",
						"padding":"1.5px 7.5px 1.5px 7.5px",})
				if not block and tag in ["<parameter>", "<attribute>"]:
					styles = utils.__append_dict__(old=styles, overwrite=True, new={
						"font_style":"italic",})
				# append tagless word.
				if styles == {}:
					_words_.append(word)
				# append tag word.
				else:
					_words_.append(self.__create_word__(word=word["value"], styles=styles))
			else:
				_words_.append(word)
		return _words_
	def __assign_element_ids__(self, variable, key="element_id"):
		if isinstance(variable, list):
			dictionary = {}
			for i in variable: dictionary[i[key]] = i
			variable = dictionary
		elif isinstance(variable, dict):
			dictionary = {}
			for _, value in variable.items(): dictionary[value[key]] = value
			variable = dictionary
		else: raise ValueError(f"Unkown variable format: {variable}")
		return variable

# the documentation view.
class DocumentationView(View):
	# do not forget to add the template dir: FilePath(website.SOURCE_PATH).base(), in settings.py
	def __init__(self, 
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the views id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# the html path (optional).
		html=None,
		# enable if this view is the [/] landing page.
		landing_page=False,
		# the template data (required).
		template_data={},
		# the default chapter library.
		chapters=None,
		# insert specific colors (insert mode).
		colors={},
		# the w3bsite.Website object.
		website=None,
	):
		# defaults.
		View.__init__(self,
			base=base,
			id=id,
			url=url,
			html=html,
			landing_page=landing_page,
			template_data=template_data,
			# the view type.
			type="DocumentationView",)
		self.website = website

		# check library.
		if not Files.exists("static/media"): os.mkdir("static/media")
		if not Files.exists("static/media/docs"): os.system(f"cp -r {SOURCE_PATH}/classes/views/media/docs/ static/media/docs")

		# add template data.
		self.colors = {
			"white":"#FAFAFA",
			"light_white":"#E9F0FD",
			"grey":"#E5E5E5",
			"light_grey":"#D6D6D6",
			"dark_grey":"#424242",
			"blue":"#5A8FE6",
			"purple":"#323B83",#"#B32FCA",
			#"purple":"#9B00AA",
			"red":"#FD304E",
			"pink":"#F62B7D",
			"orange":"#FF8800",
			"green":"#006633",
			"darkest":"#1F2227",
			"darker": "#20242A",
			"dark": "#262B30",
			# background color.
			"topbar":"#FAFAFA",#1F2227", #"#FAFAFA",
			"background":"#323B83",#"#E7E9EF", #"#FAFAFA",
			"topbar_darkmode":"#1F2227",#1F2227", #"#FAFAFA",
			"background_darkmode":"#1F2227",#"#E7E9EF", #"#FAFAFA",
			# elements.
			"widgets":"#FAFAFA",
			"widgets_reversed":"#323B83",#"#1F2227",
			"widgets_darkmode":"#20242A",
			"widgets_reversed_darkmode":"#323B83",#"#1F2227",
			# text.
			"text":"#1F2227",
			"text_reversed":"#FAFAFA",
			"text_darkmode":"#FAFAFA",
			"text_reversed_darkmode":"#FAFAFA",
			# input & textareas.
			"input_txt":"#6C6B6D",
			"input_txt_reversed":"#FAFAFA",
			"input_bg":"#E9F0FD", #"#FAFAFA", 
			"input_bg_reversed":"#323B83",
			
			# buttons.
			"button_txt":"#FAFAFA",
			"button_txt_reversed":"#1F2227",
			"button_bg":"#323B83",
			"button_bg_reversed":"#FAFAFA",
			# custom colors.
			# ...
		}
		for key, value in colors.items():self.colors[key] = value
		self.template_data["COLORS"] = self.colors
		try:
			self.template_data["CHAPTERS"] = {}
		except KeyError:
			self.template_data["CHAPTERS"] = {}
		self.template_data["leftbar_width"] = "250px"

		# arguments.
		if chapters != None:
			while True:
				if len(chapters) > 0 and chapters[0] == "/": chapters = chapters[1:]
				elif len(chapters) > 0 and chapters[len(chapters)-1] == "/": chapters = chapters[:-1]
				else: break
			self.include_chapters(chapters)

		#
	def view(self, request):
		api_key = "your-api-key"
		response = self.website.users.authenticated(request)
		if response.success:
			response = self.website.users.get_api_key(email=response.email)
			if response.success: api_key = response.api_key
		template_data = dict(self.template_data)
		template_data["api_key"] = api_key
		template_data["URL"] = self.url
		template_data["CHAPTERS"] = self.chapters
		if self.website.maintenance: return self.maintenance(request)
		return render(request, f"{ALIAS}/classes/views/html/documentation_view.html", template_data)
	def include_chapters(self, chapters_library):
		if not Files.exists(chapters_library):
			raise ValueError(f"Specified DocumentationView chapters library {chapters_library} does not exist.")
		self.chapters = {}
		if os.path.isdir(chapters_library):
			chapters_library += "/"
			replaced = chapters_library.replace('/','.').replace(".py", "")
			for path in Files.Directory(chapters_library).paths():
				id = self.__generate_element_id__("chapt_")
				name = FilePath(path).name().replace(".py", "")
				if ".py" in path:
					module = importlib.import_module(f"{replaced}{name}")
					self.chapters[id] = module.chapter
		else:
			replaced = chapters_library.replace('/','.').replace(".py", "")
			name = FilePath(chapters_library).name().replace(".py", "")
			module = importlib.import_module(f"{replaced}")
			id = self.__generate_element_id__("chapt_")
			self.chapters[id] = module.chapter
	def __generate_element_id__(self, joiner=""):
		return joiner+String("").generate(length=32, digits=True, capitalize=True)
	

# initialized docs.
docs = Documentations()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import exceptions

# the database class.
class Database(Traceback):
	def __init__(self, firestore=None, path=None):
		
		# traceback.
		Traceback.__init__(self, traceback="w3bsite.Website.database")

		# checks.
		if firestore == None and path == None: raise exceptions.InvalidUsage(self.__traceback__()+" Both parameters firestore & path are None.")

		# attributes.
		self.mode = "firestore"
		self.firestore = firestore
		self.path = gfp.clean(path)
		self.file_path = self.fp = FilePath(self.path)
		if not self.fp.exists():
			Response.log(f"&ORANGE&Root permission&END& required to create database [{self.path}].")
			os.system(f"sudo mkdir {self.path}")
			Files.chmod(path=self.path, permission=700, sudo=True)
			Files.chown(path=self.path, owner=Defaults.vars.user, group=Defaults.vars.group, sudo=True)
		if firestore == None:
			self.mode = "cache"

		#

	# functions.
	def load(self, path=None):
		if path == None: return Response.error(self.__traceback__(function="load")+" Define parameter: [path].")
		if self.mode == "firestore":
			return self.firestore.load(path)
		elif self.mode == "cache":
			path = gfp.clean(f"{self.path}/{path}")
			try:
				data = Files.load(path=path, format="json")
			except Exception as e: return Response.error(str(e))
			return Response.success(f"Successfully loaded [{path}].", {
				"data":data,
			})
		else: raise exceptions.InvalidUsage(self.__traceback__(function="load")+f" Unknown mode [{self.mode}].")
	def save(self, path=None, data=None, overwrite=False):
		if path == None: return Response.error(self.__traceback__(function="save")+" Define parameter: [path].")
		if data == None: return Response.error(self.__traceback__(function="save")+" Define parameter: [data].")
		if self.mode == "firestore":
			return self.firestore.save(path, data)
		elif self.mode == "cache":
			path = gfp.clean(f"{self.path}/{path}")
			try:
				if not Files.exists(path=gfp.base(path)): Files.create(path=gfp.base(path), directory=True)
			except ValueError: a=1
			if not overwrite:
				def insert(old, new):
					for key,value in new.items():
						if isinstance(value, (dict, Dictionary)):
							if key in old:
								old[key] = insert(old[key], value)
							else:
								old[key] = value
						elif isinstance(value, (list, Array)):
							if key in old:
								for i in value:
									if i not in old[key]: old[key].append(i)
							else:
								old[key] = value
						else:
							old[key] = value
					return old
				try: old_data = Files.load(path=path, format="json")
				except: old_data = {}
				data = insert(old_data, data)
			try:
				Files.save(path=path, data=data, format="json")
			except Exception as e: return Response.error(str(e))
			return Response.success(f"Successfully saved [{path}].")
		else: raise exceptions.InvalidUsage(self.__traceback__(function="save")+f" Unknown mode [{self.mode}].")
	def delete(self, path=None, data=None):
		if path == None: return Response.error(self.__traceback__(function="delete")+" Define parameter: [path].")
		if self.mode == "firestore":
			return self.firestore.delete(path)
		elif self.mode == "cache":
			path = gfp.clean(f"{self.path}/{path}")
			try:
				Files.delete(path=path)
			except Exception as e: return Response.error(str(e))
			if Files.exists(path): return Response.error(f"Failed to delete [{path}].")
			return Response.success(f"Successfully deleted [{path}].")
		else: raise exceptions.InvalidUsage(self.__traceback__(function="delete")+f" Unknown mode [{self.mode}].")

	# join.
	def join(self, path):
		return gfp.clean(f"{self.path}/{path}")

	# get names.
	def names(self,
		# the sub path (leave None to use the root path)
		path=None,
	):

		# vars.
		names = []

		# firestore.
		if self.mode == "firestore":
			response = self.firestore.load_collection(path)
			raise ValueError(f"Still to check the response of load_collection (for names iteration): [{response}].")

		# cache.
		elif self.mode == "cache":
			for i in Directory(path=self.join(path)).paths(): names.append(gfp.name(i))

		# handler.
		return names

		#

	# representation.
	def __str__(self):
		return str(self.fp.path)
	def __repr__(self):
		return str(self)

	#

	
			


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
import os, sys, ast, json, glob, platform, subprocess, random, requests, urllib, threading, time, string, traceback

# inc imports.
import dev0s, netw0rk, syst3m, ssht00ls
from ssht00ls import ssht00ls_agent
from dev0s import *

# source.	
ALIAS = "w3bsite"
SOURCE_PATH = Defaults.source_path(__file__, back=3)
BASE = Defaults.source_path(SOURCE_PATH)
OS = Defaults.operating_system(supported=["linux", "macos"])
LOG_LEVEL = Defaults.log_level(default=0)
#Defaults.alias(alias=ALIAS, executable=SOURCE_PATH)

# network info.
NETWORK_INFO = netw0rk.network.info()
if not NETWORK_INFO["success"]: raise ValueError(NETWORK_INFO["error"])

# check lib.
lib = Directory(f"{SOURCE_PATH}/classes/deployment/lib")
if not lib.fp.exists():# or len(lib.paths()) != 7) and Environment.get("HOST") not in ["macbookpro_daan"]:
	Response.log("Installing the w3bsite library.")
	os.system(f"rm -fr /tmp/w3bsite && git clone -q https://github.com/vandenberghinc/w3bsite /tmp/w3bsite 2> /dev/null && rsync -azq /tmp/w3bsite/w3bsite/classes/deployment/lib/ {SOURCE_PATH}/classes/deployment/lib/")

# the cache.
from w3bsite.classes.cache import cache
